@@ -77,6 +77,8 @@ MediaObject::MediaObject(Backend *backend, QObject *parent)
         , m_waitingForNextSource(false)
         , m_waitingForPreviousSource(false)
         , m_skippingEOS(false)
+        , m_skipGapless(false)
+        , m_doingEOS(false)
 {
     qRegisterMetaType<GstCaps*>("GstCaps*");
     qRegisterMetaType<State>("State");
@@ -133,6 +135,7 @@ MediaObject::~MediaObject()
     if (m_pipeline) {
         GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(m_pipeline->element()));
         g_signal_handlers_disconnect_matched(bus, G_SIGNAL_MATCH_DATA, 0, 0, 0, 0, this);
+        gst_object_unref(bus);
         delete m_pipeline;
     }
     GlobalSubtitles::instance()->unregister_(this);
@@ -144,17 +147,20 @@ void MediaObject::saveState()
     if (m_resumeState)
         return;
 
-    if (m_pendingState == Phonon::PlayingState || m_pendingState == Phonon::PausedState) {
+    if (m_state == Phonon::PlayingState || m_state == Phonon::PausedState) {
         m_resumeState = true;
-        m_oldState = m_pendingState;
+        m_oldState = m_state;
         m_oldPos = getPipelinePos();
     }
 }
 
 void MediaObject::resumeState()
 {
-    if (m_resumeState)
-        QMetaObject::invokeMethod(this, "setState", Qt::QueuedConnection, Q_ARG(State, m_oldState));
+    if (m_resumeState) {
+        m_resumeState = false;
+        requestState(m_oldState);
+        seek(m_oldPos);
+    }
 }
 
 /**
@@ -228,6 +234,7 @@ void MediaObject::setTickInterval(qint32 newTickInterval)
  */
 void MediaObject::play()
 {
+    DEBUG_BLOCK;
     requestState(Phonon::PlayingState);
 }
 
@@ -249,6 +256,8 @@ Phonon::ErrorType MediaObject::errorType() const
 
 void MediaObject::setError(const QString &errorString, Phonon::ErrorType error)
 {
+    DEBUG_BLOCK;
+    debug() << errorString;
     m_errorString = errorString;
     m_error = error;
     // Perform this asynchronously because this is also called from within the pipeline's cb_error
@@ -328,7 +337,8 @@ void MediaObject::autoDetectSubtitle()
 
 void MediaObject::setNextSource(const MediaSource &source)
 {
-    qDebug() << "Got next source. Waiting for end of current.";
+    DEBUG_BLOCK;
+    debug() << "Got next source. Waiting for end of current.";
 
     m_aboutToFinishLock.lock();
 
@@ -366,12 +376,14 @@ void MediaObject::setSource(const MediaSource &source)
     if (!isValid())
         return;
 
+    DEBUG_BLOCK;
+
     if (source.type() == Phonon::MediaSource::Invalid) {
         qWarning("Trying to set an invalid MediaSource -> ignoring.");
         return;
     }
 
-    qDebug() << "Setting new source";
+    debug() << "Setting new source";
     m_source = source;
     autoDetectSubtitle();
     m_pipeline->setSource(source);
@@ -382,26 +394,33 @@ void MediaObject::setSource(const MediaSource &source)
 // Called when we are ready to leave the loading state
 void MediaObject::loadingComplete()
 {
+    DEBUG_BLOCK;
     link();
 }
 
 void MediaObject::getSubtitleInfo(int stream)
 {
-    int text_number = 0;
-    g_object_get(G_OBJECT(m_pipeline->element()), "n-text", &text_number, NULL);
-    if (text_number)
-        GlobalSubtitles::instance()->add(this, -1, "Disable", "");
-    for (gint i = 0; i < text_number; ++i) {
-        GstTagList *tags = NULL;
+    gint spuCount = 0; // Sub picture units.
+    g_object_get(G_OBJECT(m_pipeline->element()), "n-text", &spuCount, NULL);
+    if (spuCount)
+        GlobalSubtitles::instance()->add(this, -1, tr("Disable"), "");
+    for (gint i = 0; i < spuCount; ++i) {
+        GstTagList *tags = 0;
         g_signal_emit_by_name (G_OBJECT(m_pipeline->element()), "get-text-tags",
                                i, &tags);
 
         if (tags) {
-            gchar *tlc;
-
-            gst_tag_list_get_string (tags, GST_TAG_LANGUAGE_CODE, &tlc);
-            QString name = QString("Subtitle %1 - [%2]").arg(i).arg(tlc);
+            gchar *tagLangCode = 0;
+            gst_tag_list_get_string (tags, GST_TAG_LANGUAGE_CODE, &tagLangCode);
+            QString name;
+            if (tagLangCode)
+                name = QLatin1String(tagLangCode); // Language code is ISO -> always Latin1
+            else
+                name = tr("Unknown");
             GlobalSubtitles::instance()->add(this, i, name);
+            // tagLangCode was implicat converted to QString, so we can drop
+            // the ref.
+            g_free(tagLangCode);
         }
     }
     emit availableSubtitlesChanged();
@@ -416,11 +435,13 @@ void MediaObject::setPrefinishMark(qint32 newPrefinishMark)
 
 void MediaObject::pause()
 {
+    DEBUG_BLOCK;
     requestState(Phonon::PausedState);
 }
 
 void MediaObject::stop()
 {
+    DEBUG_BLOCK;
     requestState(Phonon::StoppedState);
 }
 
@@ -429,8 +450,10 @@ void MediaObject::seek(qint64 time)
     if (!isValid())
         return;
 
+    DEBUG_BLOCK;
+
     if (m_waitingForNextSource) {
-        qDebug() << "Seeking back within old source";
+        debug() << "Seeking back within old source";
         m_waitingForNextSource = false;
         m_waitingForPreviousSource = true;
         m_pipeline->setSource(m_source, true);
@@ -502,7 +525,7 @@ Phonon::State MediaObject::translateState(GstState state) const
             return Phonon::StoppedState;
         case GST_STATE_NULL:
             return Phonon::LoadingState;
-        case GST_STATE_VOID_PENDING: //Quiet GCC
+        case GST_STATE_VOID_PENDING: // Quiet GCC
             break;
     }
     return Phonon::ErrorState;
@@ -510,10 +533,12 @@ Phonon::State MediaObject::translateState(GstState state) const
 
 void MediaObject::handleStateChange(GstState oldState, GstState newState)
 {
+    DEBUG_BLOCK;
+
     Phonon::State prevPhononState = m_state;
     prevPhononState = translateState(oldState);
     m_state = translateState(newState);
-    qDebug() << "Moving from" << GstHelper::stateName(oldState) << prevPhononState << "to" << GstHelper::stateName(newState) << m_state;
+    debug() << "Moving from" << GstHelper::stateName(oldState) << prevPhononState << "to" << GstHelper::stateName(newState) << m_state;
     if (GST_STATE_TRANSITION(oldState, newState) == GST_STATE_CHANGE_NULL_TO_READY)
         loadingComplete();
     if (GST_STATE_TRANSITION(oldState, newState) == GST_STATE_CHANGE_READY_TO_PAUSED && m_pendingTitle != 0) {
@@ -527,18 +552,28 @@ void MediaObject::handleStateChange(GstState oldState, GstState newState)
     if (newState == GST_STATE_READY)
         emit tick(0);
 
-    emit stateChanged(m_state, prevPhononState);
+    // Avoid signal emission while processing EOS to avoid bogus UI updates.
+    if (!m_doingEOS)
+        emit stateChanged(m_state, prevPhononState);
 }
 
 void MediaObject::handleEndOfStream()
 {
+    DEBUG_BLOCK;
     if (!m_skippingEOS) {
-        emit finished();
-        m_pipeline->setState(GST_STATE_READY);
+        debug() << "not skipping EOS";
+        m_doingEOS = true;
+        { // When working on EOS we do not want signals emitted to avoid bogus UI updates.
+            emit stateChanged(Phonon::StoppedState, m_state);
+            m_aboutToFinishWait.wakeAll();m_aboutToFinishLock.unlock();
+            m_pipeline->setState(GST_STATE_READY);
+            emit finished();
+        }
+        m_doingEOS = false;
     } else {
+        debug() << "skipping EOS";
         GstState state = m_pipeline->state();
         m_pipeline->setState(GST_STATE_READY);
-        m_pipeline->state();
         m_pipeline->setState(state);
         m_skippingEOS = false;
     }
@@ -780,6 +815,12 @@ void MediaObject::setMetaData(QMultiMap<QString, QString> newData)
 
 void MediaObject::requestState(Phonon::State state)
 {
+    DEBUG_BLOCK;
+    m_aboutToFinishLock.tryLock();
+    m_skipGapless = true;
+    m_aboutToFinishWait.wakeAll();
+    m_aboutToFinishLock.unlock();
+    debug() << state;
     switch (state) {
         case Phonon::PlayingState:
             m_pipeline->setState(GST_STATE_PLAYING);
@@ -802,16 +843,23 @@ void MediaObject::requestState(Phonon::State state)
 
 void MediaObject::handleAboutToFinish()
 {
-    qDebug() << "About to finish";
+    DEBUG_BLOCK;
+    debug() << "About to finish";
     m_aboutToFinishLock.lock();
     emit aboutToFinish();
     // Three seconds should be more than enough for any application to get their act together.
     // Any longer than that and they have bigger issues.  If Phonon does no supply a next source
     // within 3 seconds, treat as if there is no next source to come, and finish the current source.
-    if (m_aboutToFinishWait.wait(&m_aboutToFinishLock, 3000))
-        qDebug() << "Finally got a source";
-    else
-        m_skippingEOS = false;
+    if (!m_skipGapless) {
+      if (m_aboutToFinishWait.wait(&m_aboutToFinishLock, 3000)) {
+          debug() << "Finally got a source";
+      } else {
+          m_skippingEOS = false;
+      }
+    } else {
+      debug() << "Skipping gapless audio";
+    }
+    m_skipGapless = false;
     m_aboutToFinishLock.unlock();
 }
 
